use std::error::Error;

fn factorial(num: usize) -> usize {
    let mut fact = 1;
    for n in 1..=num {
        fact *= n;
    }
    fact
}

fn binomial(n: usize, k: usize) -> usize {
    if k > n {
        0
    } else {
        factorial(n) / (factorial(k) * factorial(n - k))
    }
}

pub fn nth_lex<T: Clone + Ord>(
    mut elements: Vec<T>,
    k: usize,
    nth: usize,
) -> Result<Vec<T>, Box<dyn Error>> {
    elements.sort();
    let num_elements = elements.len();
    let num_combinations = binomial(num_elements, k);
    if nth > num_combinations || k > num_elements || nth == 0 || k == 0 {
        return Err(Box::from("Out of bounds"));
    }
    let mut i = 0;
    let mut remaining_k = k;
    let mut comb = Vec::new();
    let mut remainder = nth - 1;
    while remaining_k > 0 {
        // Count the number of combinations that start with elements[i]
        // example with n = 5, k = 2
        // nth <= 4 select first
        // nth <= 7 select second
        // nth <= 9 select third
        // nth == 10 select fourth
        let count = binomial(num_elements - i - 1, remaining_k - 1);
        if remainder < count {
            // If the nth combination is within the count, pick this element
            comb.push(elements[i].clone());
            remaining_k -= 1;
        } else {
            remainder -= count;
        }
        i += 1;
    }
    Ok(comb)
}

#[derive(Clone)]
pub struct Combinator<T: Clone + Ord> {
    pub current: Vec<T>,
    pub k: usize,
    idx: usize,
}

impl<T: Clone + Ord> Combinator<T> {
    pub fn new(elements: Vec<T>, k: usize) -> Result<Combinator<T>, Box<dyn Error>> {
        if k > elements.len() || k == 0 {
            return Err(Box::from("Out of bounds"));
        }
        Ok(Self {
            current: elements,
            k,
            idx: 0,
        })
    }
}

impl<T: Clone + Ord> Iterator for Combinator<T> {
    type Item = Vec<T>;

    fn next(&mut self) -> Option<Self::Item> {
        let num_elements = self.current.len();
        let num_combinations = binomial(num_elements, self.k);
        if self.idx == num_combinations {
            return None;
        }
        let mut i = 0;
        let mut remaining_k = self.k;
        let mut comb = Vec::new();
        let mut remainder = self.idx - 1;
        while remaining_k > 0 {
            // Count the number of combinations that start with elements[i]
            // example with n = 5, k = 2
            // nth <= 4 select first
            // nth <= 7 select second
            // nth <= 9 select third
            // nth == 10 select fourth
            let count = binomial(num_elements - i - 1, remaining_k - 1);
            if remainder < count {
                // If the nth combination is within the count, pick this element
                comb.push(self.current[i].clone());
                remaining_k -= 1;
            } else {
                remainder -= count;
            }
            i += 1;
        }
        self.idx += 1;
        self.current = comb;
        Some(self.current.clone())
    }
}

#[cfg(test)]
mod test {
    use crate::combination::nth_lex;

    const SMALL: [i8; 5] = [1, 2, 3, 4, 5];

    // Nth Lex

    #[test]
    fn nth_lex_zero_small() {
        let small = SMALL.to_vec();
        assert!(nth_lex(small, 0, 1).is_err());
    }

    #[test]
    fn nth_lex_more_small() {
        let small = SMALL.to_vec();
        assert!(nth_lex(small, 6, 1).is_err());
    }

    #[test]
    fn nth_lex_all_small() {
        let small = SMALL.to_vec();
        assert!(nth_lex(small.clone(), 5, 1).unwrap() == vec![1, 2, 3, 4, 5]);
        assert!(nth_lex(small, 5, 2).is_err());
    }

    #[test]
    fn nth_lex_some_small() {
        let small = SMALL.to_vec();
        assert_eq!(nth_lex(small.clone(), 2, 1).unwrap(), vec![1, 2]);
        assert_eq!(nth_lex(small.clone(), 2, 2).unwrap(), vec![1, 3]);
        assert_eq!(nth_lex(small.clone(), 2, 3).unwrap(), vec![1, 4]);
        assert_eq!(nth_lex(small.clone(), 2, 4).unwrap(), vec![1, 5]);
        assert_eq!(nth_lex(small.clone(), 2, 5).unwrap(), vec![2, 3]);
        assert_eq!(nth_lex(small.clone(), 2, 6).unwrap(), vec![2, 4]);
        assert_eq!(nth_lex(small.clone(), 2, 7).unwrap(), vec![2, 5]);
        assert_eq!(nth_lex(small.clone(), 2, 8).unwrap(), vec![3, 4]);
        assert_eq!(nth_lex(small.clone(), 2, 9).unwrap(), vec![3, 5]);
        assert_eq!(nth_lex(small.clone(), 2, 10).unwrap(), vec![4, 5]);
        assert!(nth_lex(small, 2, 11).is_err());
    }

    #[test]
    fn nth_lex_one_small() {
        let small = SMALL.to_vec();
        assert_eq!(nth_lex(small.clone(), 1, 1).unwrap(), vec![1]);
        assert_eq!(nth_lex(small.clone(), 1, 2).unwrap(), vec![2]);
        assert_eq!(nth_lex(small.clone(), 1, 3).unwrap(), vec![3]);
        assert_eq!(nth_lex(small.clone(), 1, 4).unwrap(), vec![4]);
        assert_eq!(nth_lex(small.clone(), 1, 5).unwrap(), vec![5]);
        assert!(nth_lex(small, 1, 6).is_err());
    }

    #[test]
    fn nth_lex_all_big() {
        let big: Vec<i32> = (0..100).collect();
        assert_eq!(
            nth_lex(big.clone(), 100, 1).unwrap(),
            (0..100).collect::<Vec<i32>>()
        );
        assert!(nth_lex(big, 100, 2).is_err());
    }

    #[test]
    fn nth_lex_some_big() {
        let big: Vec<i32> = (0..100).collect();
        assert_eq!(
            nth_lex(big.clone(), 50, 1).unwrap(),
            (0..50).collect::<Vec<i32>>()
        );
    }

    #[test]
    fn nth_lex_one_big() {
        let big: Vec<i32> = (0..100).collect();
        assert_eq!(nth_lex(big.clone(), 1, 1).unwrap(), vec![0]);
        assert_eq!(nth_lex(big.clone(), 1, 2).unwrap(), vec![1]);
        assert_eq!(nth_lex(big.clone(), 1, 3).unwrap(), vec![2]);
        assert_eq!(nth_lex(big.clone(), 1, 4).unwrap(), vec![3]);
    }

    // Combinator. Needs fixing

    //#[test]
    //fn comb_zero_small() {
    //    let small = SMALL.to_vec();
    //    let comb = Combinator::new(small, 0);
    //    assert!(comb.is_err());
    //}

    //#[test]
    //fn comb_more_small() {
    //    let small = SMALL.to_vec();
    //    let comb = Combinator::new(small, 6);
    //    assert!(comb.is_err());
    //}

    //#[test]
    //fn comb_all_small() {
    //    let small = SMALL.to_vec();
    //    let mut comb = Combinator::new(small, 5).unwrap();
    //    assert!(comb.next() == Some(vec![1, 2, 3, 4, 5]));
    //    assert!(comb.next().is_none());
    //}

    //#[test]
    //fn comb_some_small() {
    //    let small = SMALL.to_vec();
    //    let mut comb = Combinator::new(small, 2).unwrap();
    //    assert_eq!(comb.next(), Some(vec![1, 2]));
    //    assert_eq!(comb.next(), Some(vec![1, 3]));
    //    assert_eq!(comb.next(), Some(vec![1, 4]));
    //    assert_eq!(comb.next(), Some(vec![1, 5]));
    //    assert_eq!(comb.next(), Some(vec![2, 3]));
    //    assert_eq!(comb.next(), Some(vec![2, 4]));
    //    assert_eq!(comb.next(), Some(vec![2, 5]));
    //    assert_eq!(comb.next(), Some(vec![3, 4]));
    //    assert_eq!(comb.next(), Some(vec![3, 5]));
    //    assert_eq!(comb.next(), Some(vec![4, 5]));
    //    assert!(comb.next().is_none());
    //}

    //#[test]
    //fn comb_one_small() {
    //    let small = SMALL.to_vec();
    //    let mut comb = Combinator::new(small, 1).unwrap();
    //    assert_eq!(comb.next(), Some(vec![1]));
    //    assert_eq!(comb.next(), Some(vec![2]));
    //    assert_eq!(comb.next(), Some(vec![3]));
    //    assert_eq!(comb.next(), Some(vec![4]));
    //    assert_eq!(comb.next(), Some(vec![5]));
    //    assert!(comb.next().is_none());
    //}
}
