use std::{
    collections::{HashMap, HashSet},
    error::Error,
};

use crate::number::{Number, Sign};

#[derive(Debug, PartialEq, Eq)]
pub struct Fraction {
    pub numerator: Number,
    pub denominator: Number,
    pub sign: Sign,
}

pub struct Rational {
    pub decimal: Number,
    pub mantissa: Number,
    pub recurring: Number,
}

impl Fraction {
    pub fn new(mut numerator: Number, mut denominator: Number) -> Result<Self, Box<dyn Error>> {
        if denominator == 0.into() {
            return Err(Box::from("Division by 0"));
        }
        let sign = match numerator.sign {
            Sign::Positif => match denominator.sign {
                Sign::Positif => Sign::Positif,
                Sign::Negatif => Sign::Negatif,
            },
            Sign::Negatif => match denominator.sign {
                Sign::Positif => Sign::Negatif,
                Sign::Negatif => Sign::Positif,
            },
        };
        numerator.sign = Sign::Positif;
        denominator.sign = Sign::Positif;
        let mut f = Fraction {
            numerator,
            denominator,
            sign,
        };
        f.reduce();
        Ok(f)
    }

    pub fn reduce(&mut self) {
        let gcd = Number::gcd(self.numerator.clone(), self.denominator.clone());
        if gcd == self.denominator.clone() {
            return;
        }
        self.numerator = self.numerator.clone() / gcd.clone();
        self.denominator = self.denominator.clone() / gcd.clone();
    }

    pub fn get_full_mantissa(fraction: &Fraction) -> Option<Number> {
        let mut div_memory = HashMap::new();
        let mut rem_memory = HashSet::new();
        rem_memory.insert(1.into());
        let mut div_with = fraction.numerator.clone();
        let mut rational = Vec::new();
        if div_with < fraction.denominator {
            div_with = div_with * 10;
        }
        loop {
            while div_with < fraction.denominator {
                rational.push(0);
                div_memory
                    .entry(0)
                    .and_modify(|value| *value += 1)
                    .or_insert(1);
                div_with = div_with * 10;
                rem_memory.insert(div_with.clone());
            }
            let (div, rem) =
                Number::div_with_rem(div_with.clone(), fraction.denominator.clone()).unwrap();
            if rem == 0.into() {
                return None;
            }
            let next_digit = div.digits[0];
            if div_memory.contains_key(&next_digit) && rem_memory.contains(&rem) && next_digit != 0
            {
                let mut digits = vec![];
                if div_memory.values().min() != div_memory.values().max() {
                    rational.push(div.digits[0]);
                    digits = rational;
                    let (l, r) = digits.split_at(digits.len() / 2);
                    if l == r {
                        digits = l.to_vec();
                    }
                } else {
                    let idx = rational
                        .iter()
                        .position(|digit| digit == &div.digits[0])
                        .unwrap();
                    if idx == rational.len() - 1 {
                        digits.push(div.digits[0]);
                    } else {
                        for digit in &rational[idx..] {
                            if digits.contains(digit) {
                                break;
                            }
                            digits.push(*digit);
                        }
                    }
                }
                return Some(Number {
                    digits,
                    sign: Sign::Positif,
                });
            }
            div_memory
                .entry(next_digit)
                .and_modify(|count| *count += 1)
                .or_insert(1);
            rem_memory.insert(rem.clone());
            rational.push(div.digits[0]);
            div_with = rem * 10;
        }
    }
}

#[cfg(test)]
mod test {
    use crate::number::Sign;

    use super::Fraction;

    #[test]
    fn test_reduce() {
        let f = Fraction::new(120.into(), 90.into()).unwrap();
        assert_eq!(
            f,
            Fraction {
                numerator: 4.into(),
                denominator: 3.into(),
                sign: Sign::Positif
            }
        );
    }
}
