mod add;
mod div;
mod from;
mod mul;
mod ord;
mod sub;

use std::fmt::{Display, Formatter};

#[derive(Copy, Clone, Debug, Eq, PartialEq, Hash)]
pub enum Sign {
    Positif,
    Negatif,
}

#[derive(Clone, Debug, Eq, PartialEq, Hash)]
pub struct Number {
    pub digits: Vec<i8>,
    pub sign: Sign,
}

impl Number {
    pub fn get_digit(n: isize, pos: usize) -> isize {
        let modulo = 10isize.pow(pos as u32 + 1);
        let divisor = modulo / 10;
        (n % modulo) / divisor
    }

    pub const fn byte_to_digit(b: u8) -> i8 {
        // wrapping_sub('0' as u32) same as - 48 but less magical
        (b as i8).wrapping_sub('0' as i8)
    }

    pub fn pow(self, n: u32) -> Self {
        let mut result = self.clone();
        if ((self.digits.len() * 8) as u32) < isize::BITS {
            let number = isize::try_from(self).unwrap();
            for _i in 1..n {
                result = result * number;
            }
            result
        } else {
            let number = self.clone();
            for _i in 1..n {
                result *= number.clone();
            }
            result
        }
    }

    pub fn ten_pow(self, n: u32) -> Self {
        Number {
            digits: [vec![0; n as usize], self.digits].concat(),
            sign: self.sign,
        }
    }

    pub fn fact(self) -> Self {
        let mut fact = Number::from(1);
        let mut n = Number::from(1);
        while n <= self {
            fact *= n.clone();
            n += 1.into();
        }
        fact
    }

    pub fn fib(n: u64) -> Self {
        let mut last_two = (Self::from(1), Self::from(1));
        let mut iteration = 1;
        while iteration < n {
            last_two = (last_two.1.clone(), last_two.0 + last_two.1);
            iteration += 1;
        }

        last_two.0
    }

    pub fn abs(self) -> Self {
        Number {
            digits: self.digits,
            sign: Sign::Positif,
        }
    }
}

impl Display for Number {
    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
        let number_string = self
            .digits
            .iter()
            .rev()
            .map(|&digit| digit.to_string())
            .collect::<Vec<String>>()
            .join("");
        match self.sign {
            Sign::Positif => write!(f, "{number_string}"),
            Sign::Negatif => write!(f, "-{number_string}"),
        }
    }
}

#[cfg(test)]
mod number_tests {
    use crate::number::Number;

    #[test]
    fn test_get_digit() {
        let num = 12345;
        let digit_1 = Number::get_digit(num, 0);
        let digit_2 = Number::get_digit(num, 1);
        let digit_3 = Number::get_digit(num, 2);
        let digit_4 = Number::get_digit(num, 3);
        let digit_5 = Number::get_digit(num, 4);
        assert_eq!(digit_1, 5);
        assert_eq!(digit_2, 4);
        assert_eq!(digit_3, 3);
        assert_eq!(digit_4, 2);
        assert_eq!(digit_5, 1);
    }

    #[test]
    fn test_pow() {
        let num = Number::from(2);
        assert_eq!(Number::from(16), num.pow(4));
    }

    #[test]
    fn test_ten_pow_zero() {
        let num = Number::from(2);
        assert_eq!(Number::from(2), num.ten_pow(0));
    }

    #[test]
    fn test_ten_pow_one() {
        let num = Number::from(2);
        assert_eq!(Number::from(20), num.ten_pow(1));
    }

    #[test]
    fn test_ten_pow_two() {
        let num = Number::from(2);
        assert_eq!(Number::from(200), num.ten_pow(2));
    }
}
