use core::panic;
use std::{
    error::Error,
    ops::{Div, DivAssign, Rem, RemAssign},
};

use super::{Number, Sign};

impl Number {
    pub fn div_with_rem(
        dividend: Number,
        divisor: Number,
    ) -> Result<(Number, Number), Box<dyn Error>> {
        if divisor == 0.into() {
            return Err(Box::from("Division by 0"));
        }
        let dividend_len = dividend.digits.len();
        let divisor_len = divisor.digits.len();
        let sign = match (dividend.sign, divisor.sign) {
            (Sign::Positif, Sign::Positif) | (Sign::Negatif, Sign::Negatif) => Sign::Positif,
            (Sign::Positif, Sign::Negatif) | (Sign::Negatif, Sign::Positif) => Sign::Negatif,
        };
        let mut dividend = Number {
            digits: dividend.digits,
            sign: Sign::Positif,
        };
        let divisor = Number {
            digits: divisor.digits,
            sign: Sign::Positif,
        };
        match (sign, dividend.cmp(&divisor)) {
            (_, std::cmp::Ordering::Less) => return Ok((Number::from(0), dividend)),
            (Sign::Positif, std::cmp::Ordering::Equal) => {
                return Ok((Number::from(1), Number::from(0)))
            }
            (Sign::Negatif, std::cmp::Ordering::Equal) => {
                return Ok((Number::from(-1), Number::from(0)))
            }
            (_, std::cmp::Ordering::Greater) => (),
        }
        let mut quotient = vec![];
        let mut remainder = Number {
            digits: dividend
                .digits
                .drain((dividend_len - divisor_len)..)
                .collect(),
            sign: Sign::Positif,
        };

        loop {
            while remainder < divisor && !dividend.digits.is_empty() {
                remainder.digits.insert(0, dividend.digits.pop().unwrap());
                quotient.push(0);
            }
            let mut factor = 0;
            loop {
                if remainder.digits.iter().max().unwrap() == &0 {
                    quotient.push(0);
                    break;
                }
                let temp_remainder = remainder.clone() - divisor.clone();
                if temp_remainder.sign == Sign::Negatif {
                    quotient.push(factor);
                    break;
                }
                factor += 1;
                if temp_remainder == 0.into() {
                    remainder = temp_remainder;
                    quotient.push(factor);
                    break;
                }
                remainder = temp_remainder;
            }
            if dividend.digits.is_empty() {
                break;
            }
            remainder.digits.insert(0, dividend.digits.pop().unwrap());
        }
        for digit in quotient.clone() {
            if digit != 0 || quotient.len() == 1 {
                break;
            }
            quotient.remove(0);
        }
        quotient.reverse();
        let res = Number {
            digits: quotient,
            sign,
        };
        let mut rem_digits = remainder.clone().digits;
        rem_digits.reverse();
        for digit in rem_digits {
            if digit != 0 || remainder.digits.len() == 1 {
                break;
            }
            remainder.digits.pop();
        }
        Ok((res, remainder))
    }

    pub fn gcd(mut a: Number, mut b: Number) -> Number {
        loop {
            let t = b.clone();
            b = a % b;
            a = t;
            if b == Number::from(0) {
                return a;
            }
        }
    }
}

impl Div for Number {
    type Output = Self;

    fn div(self, rhs: Self) -> Self::Output {
        Self::div_with_rem(self, rhs).unwrap().0
    }
}

impl DivAssign for Number {
    fn div_assign(&mut self, rhs: Self) {
        match Self::div_with_rem(self.clone(), rhs) {
            Ok((new, _)) => {
                self.digits = new.digits;
                self.sign = new.sign;
            }
            Err(e) => panic!("{e}"),
        }
    }
}

impl Rem for Number {
    type Output = Self;

    fn rem(self, rhs: Self) -> Self::Output {
        Self::div_with_rem(self, rhs).unwrap().1
    }
}

impl RemAssign for Number {
    fn rem_assign(&mut self, rhs: Self) {
        match Self::div_with_rem(self.clone(), rhs) {
            Ok((_, new)) => {
                self.digits = new.digits;
                self.sign = new.sign;
            }
            Err(e) => panic!("{e}"),
        }
    }
}

#[cfg(test)]
mod test_number_div_rem {
    use core::panic;

    use crate::number::Number;

    #[test]
    fn zero_denominator() {
        let a = Number::from(9);
        let b = Number::from(0);
        if Number::div_with_rem(a, b).is_ok() {
            panic!("Division by 0 not allowed");
        }
    }

    #[test]
    fn div_rem_positif() {
        let a = Number::from(9);
        let b = Number::from(9);
        let res = (Number::from(1), Number::from(0));
        assert_eq!(res, Number::div_with_rem(a, b).unwrap());
    }

    #[test]
    fn div_rem_negatif() {
        let a = Number::from(-9);
        let b = Number::from(-9);
        let res = (Number::from(1), Number::from(0));
        assert_eq!(res, Number::div_with_rem(a, b).unwrap());
    }

    #[test]
    fn div_rem_positif_with_negatif() {
        let a = Number::from(9);
        let b = Number::from(-9);
        let res = (Number::from(-1), Number::from(0));
        assert_eq!(res, Number::div_with_rem(a, b).unwrap());
    }

    #[test]
    fn div_rem_negatif_with_positif() {
        let a = Number::from(-9);
        let b = Number::from(9);
        let res = (Number::from(-1), Number::from(0));
        assert_eq!(res, Number::div_with_rem(a, b).unwrap());
    }

    #[test]
    fn div_rem_eleven_x_ten() {
        let a = Number::from(11);
        let b = Number::from(10);
        let res = (Number::from(1), Number::from(1));
        assert_eq!(res, Number::div_with_rem(a, b).unwrap());
    }

    #[test]
    fn div_rem_eleven_x_eleven() {
        let a = Number::from(11);
        let b = Number::from(11);
        let res = (Number::from(1), Number::from(0));
        assert_eq!(res, Number::div_with_rem(a, b).unwrap());
    }

    #[test]
    fn div_rem_ten_x_eleven() {
        let a = Number::from(10);
        let b = Number::from(11);
        let res = (Number::from(0), Number::from(10));
        assert_eq!(res, Number::div_with_rem(a, b).unwrap());
    }

    #[test]
    fn div_rem_big_1() {
        let a = Number::from("123");
        let b = Number::from("32");
        let res = (Number::from(3), Number::from(27));
        assert_eq!(res, Number::div_with_rem(a, b).unwrap());
    }

    #[test]
    fn div_rem_big_2() {
        let a = Number::from("123456789");
        let b = Number::from("123");
        let res = (Number::from(1003713), Number::from(90));
        assert_eq!(res, Number::div_with_rem(a, b).unwrap());
    }

    #[test]
    fn div_rem_big_3() {
        let a = Number::from("1010101010");
        let b = Number::from("1010");
        let res = (Number::from(1000100), Number::from(10));
        assert_eq!(res, Number::div_with_rem(a, b).unwrap());
    }

    #[test]
    fn div_rem_small() {
        let a = Number::from(10);
        let b = Number::from(1);
        let res = (Number::from(10), Number::from(0));
        assert_eq!(res, Number::div_with_rem(a, b).unwrap());
    }
}
