use std::error::Error;

fn factorial(num: usize) -> usize {
    let mut fact = 1;
    for n in 1..=num {
        fact *= n;
    }
    fact
}

/// Explanation
///
/// there are 10! possible permutations
/// for each first number there are 9!, for each first 2 numbers 8!, etc.
/// we check how many times we have 9! permutations before we're over 1_000_000
/// aka. 1000000 / 9!
/// we take the remainder and check how many times we have 8! before we?re over it
/// (1000000 % 9!) 8!
/// etc.
/// every iteration we remove the digit by the idx from the original permutation
/// we only check for 999999 permutations because we already have the first one
///
pub fn nth_lex<T: Clone + Ord>(mut digits: Vec<T>, nth: usize) -> Result<Vec<T>, Box<dyn Error>> {
    digits.sort();
    if nth == 1 {
        return Ok(digits);
    }
    if nth > factorial(digits.len()) || nth == 0 {
        return Err(Box::from("Out of bounds"));
    }
    let mut perm = Vec::new();
    let num_unique_digits = digits.len();
    let mut remainder = nth - 1;
    for idx in 1..=digits.len() {
        let permutations = remainder / factorial(num_unique_digits - idx);
        remainder %= factorial(num_unique_digits - idx);
        perm.push(digits[permutations].clone());
        digits.remove(permutations);
    }
    Ok(perm)
}

#[derive(Clone)]
pub struct Permutator<T: Clone + Ord> {
    pub elements: Vec<T>,
    idx: usize,
}

impl<T: Clone + Ord> Permutator<T> {
    pub fn new(elements: Vec<T>) -> Self {
        Self { elements, idx: 0 }
    }
}

impl<T: Clone + Ord + std::fmt::Display> Iterator for Permutator<T> {
    type Item = Vec<T>;
    /// Returns the next permutation and changes the current permutation to it
    /// This operation wraps around
    fn next(&mut self) -> Option<Self::Item> {
        if self.elements.is_empty() {
            return None;
        }
        if self.idx == 0 {
            self.idx += 1;
            return Some(self.elements.clone());
        }
        let num_unique_digits = self.elements.len();
        if self.idx == factorial(num_unique_digits) {
            return None;
        }
        let mut digits = self.elements.clone();
        let mut perm = Vec::new();
        let mut remainder = self.idx;
        for idx in 1..=num_unique_digits {
            let permutations = remainder / factorial(num_unique_digits - idx);
            remainder %= factorial(num_unique_digits - idx);
            perm.push(digits.remove(permutations));
        }
        self.idx += 1;
        Some(perm)
    }
}

#[cfg(test)]
mod test {
    use crate::permutation::{nth_lex, Permutator};

    const SMALL: [i8; 3] = [1, 2, 3];

    // Nth Lex

    #[test]
    fn nth_lex_zero_small() {
        let small = SMALL.to_vec();
        assert!(nth_lex(small, 0).is_err());
    }

    #[test]
    fn nth_lex_small() {
        let small = SMALL.to_vec();
        assert_eq!(nth_lex(small.clone(), 1).unwrap(), vec![1, 2, 3]);
        assert_eq!(nth_lex(small.clone(), 2).unwrap(), vec![1, 3, 2]);
        assert_eq!(nth_lex(small.clone(), 3).unwrap(), vec![2, 1, 3]);
        assert_eq!(nth_lex(small.clone(), 4).unwrap(), vec![2, 3, 1]);
        assert_eq!(nth_lex(small.clone(), 5).unwrap(), vec![3, 1, 2]);
        assert_eq!(nth_lex(small.clone(), 6).unwrap(), vec![3, 2, 1]);
        assert!(nth_lex(small.clone(), 7).is_err());
    }

    // Permutator
    #[test]
    fn perm_small() {
        let small = SMALL.to_vec();
        let mut comb = Permutator::new(small);
        assert_eq!(comb.next(), Some(vec![1, 2, 3]));
        assert_eq!(comb.next(), Some(vec![1, 3, 2]));
        assert_eq!(comb.next(), Some(vec![2, 1, 3]));
        assert_eq!(comb.next(), Some(vec![2, 3, 1]));
        assert_eq!(comb.next(), Some(vec![3, 1, 2]));
        assert_eq!(comb.next(), Some(vec![3, 2, 1]));
        assert_eq!(comb.next(), None);
    }
}
